## Directory overview

### Configuration files

The folder contains scripts to set up the environment before launching the *ecWAM* model, as well as the `SBATCH` script files. It currently holds:
- *LEONARDO*:
    - `setup-env-Leo.sh`: to run with `MPI+OpenMP`. Notice that the executables within the container are located in `/opt/ecwam/install/bin/`
    - `setup-env-Leo.gpu.sh`: to run on GPU with `OpenACC`. Notice that the executables within the container are located in `/opt/ecwam_gpu/install/bin/`, and the Singularity commands are passed with the `--nv` flag to allow mounting the host GPU drivers at runtime
- *GALILEO*: similarly to the above we have a `setup-env-Gal.sh` and a `setup-env-Gal.gpu.sh`configuration file
- *LOCALLY*: a `setup-env-local.sh` configuration file is provided to run the code on a local machine


For the *GALILEO* and *LEONARDO* configuration files we load the necessary modules for `openmpi`, `cuda` and `singularity`. Notice that, for *LEONARDO*, we do not need to load any `Singularity` module as `SingularityPRO@3.9.9` has been recently installed on the machine.


In all cases, we set the `$ECWAM_PROJECT_NAME`, `$PREPROC`, `$PRESET`, `$CHIEF` and `$MODEL` environment variables. *Originally* these variables would point to **binaries** of the model, while *now* they invoke the execution of the same binaries (hosted in the container) via `Singularity`.

Contextually, we set the environment variables
- `ECWAM_RUN_DIR=$PWD/bind_folders/run_dir`
- `ECWAM_DATA_DIR=$PWD/bind_folders/data_dir`
- `ECWAM_CACHE_PATH=$PWD/bind_folders/cache_dir`
used to specify the location of the `run`, `data` and `cache` directories. 

 
Finally, we set the `$SINGULARITY_BIND` variable, holding the `src:dest` of folders to be mounted whence an instance of the container is started. In particular
- we bind the `run_dir/`, `data_dir/` and `cache_dir/` folders to allow writing at runtime during the execution of *ecWAM*
- in the clusters, we also mount the `$TMPDIR` (or `$SCRATCH`) folder for MPI and the work area (e.g. `/g100_work`)
- as well as the **PMIx** libraries to allow the code to run on multiple nodes



### Compatibility

The containerized *ecWAM* model, should be compiled with a `CUDA` version compatible with that installed in the host system. At the moment, 2023/11/24, the specifics of the CINECA clusters are
+ *LEONARDO*: `CUDA` version `11.8`, driver version `520.61.05`
+ *GALILEO*: `CUDA` version `11.4`, driver version `470.42.01`

Depending on the chosen compiler (have a look at the `/ecwam-container/definitions` subfolder and the `README.md` therein), we provide versions `11.0`, `11.8`, `12.0` and `12.2` of `CUDA`. In virtue of the above specs and also of the [CUDA compatibility table](https://docs.nvidia.com/deploy/cuda-compatibility/), one should use
+ *On LEONARDO*: a container with either `CUDA@11.0` or `CUDA@11.8`
+ *On GALILEO*: a container with `CUDA@11.0`



### Running on Leonardo

In this folder we also include the `run_pre.sh` and `sbatch-Leo.sh` scripts for running the containerized model in Leonardo. 

***WARNING***: In the *LEONARDO* cluster, the `$TMPDIR` variable is present in the login nodes but **not** in the compute nodes; on the other side, if present, the `$TMPDIR` folder must be binded to the container image, as it will be used by `openmpi` at runtime. By default, the configuration files we provide do not bind the `$TMPDIR` variables. This means that running `run_pre.sh` (or the model itself) on the login nodes will require first to source the desired configuration file and hence to bind `$TMPDIR` manually as:

    . setup-env-Leo.gpu.sh 
    export $SINGULARITY_BIND=$TMPDIR,$SINGULARITY_BIND

A brief description of the two scripts: 
- `run_pre.sh`: executes *serially* the `preproc` and `preset` phases. It should be called passing the `realpath` of the container image as in

        run_pre.sh      # optional: /path/to/my/ecwam/sif
    
    it otherwise assumes it to be located at `/ecwam-container/images/container_ecwam.sif`. The script first sources the configuration file `setup-env-Leo.gpu.sh` and hence executes `/ecwam-container/scripts/ecwam_run_preproc.sh` and `/ecwam-container/scripts/ecwam_run_preset.sh`. 
- `sbatch-Leo.sh`: runs the model on the *LEONARDO* cluster. For convenience, it is invoked passing the total number of nodes `$nnodes` and tasks `$ntasks` via the correspondig `sbatch` flags `-N` and `-n`; in order to exploit the 4 GPUs per node of Leonardo, one should set `$ntasks=4*$nnodes`.

        sbatch -t "00:10:00" -N $nnodes -n $ntasks sbatch-Leo.sh $nnodes /path/to/my/ecwam/sif
        sbatch -t "00:10:00" -N 3 -n 12 sbatch-Leo.sh 3        #e.g.

    The variable `$nnodes` is passed to the `sbatch-Leo.sh` script as well since it is needed in setting the `LAUNCH` variable. The script defines the `#SBATCH` options to be used for the job, namely:
    - `#SBATCH --ntasks-per-node=4`: 4 MPI tasks per node, one per GPU
    - `#SBATCH --cpus-per-task=8`: 8 cores per MPI tasks, saturating the 8*4=32 cores on the *LEONARDO* CPUs
    - `#SBATCH --gres=gpu:4`: 4 GPUs per node are requested
    - `#SBATCH --exclusive`: reserves the whole node
    - `#SBATCH -p boost_usr_prod`: to run in the GPU partition in *LEONARDO*
    - `##SBATCH -q boost_qos_lprod`: ad hoc queue for bigger jobs, the number of nodes reserved must be in the range `[33,256]`
    - `#SBATCH -A DE360_leonardo`: cluster account used to run the 
    - `export OMP_NUM_THREADS=8`: we enforce 8 threads per MPI task

    As above, the path to the container image should also be passed as second variable to `sbatch-Leo.sh`, or the default path `/ecwam-container/images/container_ecwam.sif` is assumed. The script sources the configuration file `setup-env-Leo.gpu.sh`, sets the variables 
        
            export LAUNCH="mpirun -np ${TASKS}"
    
    and finally executes `/ecwam-container/scripts/ecwam_run_model.sh`. 

The model can also be run with `salloc` and `srun` specifying the corresponding `SBATCH` flags on the command line, sourcing the configuration file beforehand and setting the `LAUNCH` variable. For example:

    salloc -t 03:00:00 --nodes=32  --ntasks-per-node=4 --ntasks=128 --exclusive --gres=gpu:4 -p boost_usr_prod -A DE360_leonardo
    . setup-env-Leo.gpu.sh                                              # optionally specify: /path/to/my/ecwam/sif 
    export OMP_NUM_THREADS=8
    export LAUNCH="srun --nodes=12 --ntasks-per-node=4 --ntasks=48"     # eg running on a subset of the salloc
    ../ecwam_scripts/ecwam_run_model.sh

The above assumes the `$ECWAM_RUN_DIR`, `$ECWAM_DATA_DIR` and `$EWAM_CACHE_PATH` environment variables to have been set. Otherwise, one should specify them in the command line, e.g.

    ../ecwam_scripts/ecwam_run_model.sh --run-dir=$PWD --config=/path/to/config.yml
