#!/bin/bash


# Get the absolute path of the script directory
script_dir="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)"
base_path=$(cd "$script_dir/.." && pwd -P)


# Default values
NVHPC_VERSION="23.1"
CUDA_VERSION="11.8"
CUDA_FLAG_SET=false
REPONAME="$base_path"
MULTI=false



# Function to validate CUDA version based on NVHPC version
validate_cuda_version() {
  local nvhpc_version="$1"
  local cuda_version="$2"

  if [ "$nvhpc_version" == "23.1" ]; then
    case "$cuda_version" in
      11.0 | 11.8 | 12.0)
        ;;
      *)
        echo "Invalid CUDA version $cuda_version for NVHPC version 23.1. Allowed values are 11.0, 11.8, and 12.0."
        exit 1
        ;;
    esac
  elif [ "$nvhpc_version" == "23.9" ]; then
    case "$cuda_version" in
      11.0 | 11.8 | 12.2)
        ;;
      *)
        echo "Invalid CUDA version $cuda_version for NVHPC version 23.9. Allowed values are 11.0, 11.8, and 12.2."
        exit 1
        ;;
    esac
  fi
}


# Parse command-line options
if [ "$#" -eq 0 ]; then
  echo "No options provided. Using default values."
else
  while [ "$#" -gt 0 ]; do
    case "$1" in
      --nvhpc=*)
        NVHPC_VERSION="${1#*=}"
        if [ "$NVHPC_VERSION" != "23.1" ] && [ "$NVHPC_VERSION" != "23.9" ]; then
          echo "Invalid value for --nvhpc: ${1#*=}. Allowed values are 23.1 and 23.9." >&2
          exit 1
        fi
        ;;
      --cuda=*)
        if [ "$MULTI" == true ]; then
          echo "Ignoring --cuda flag when multi is set to true. Creating definition file for a CUDA Multi distribution."
          CUDA_VERSION="MULTI"
        else
          CUDA_VERSION="${1#*=}"
          CUDA_FLAG_SET=true
        fi
        ;;
      --multi=*)
        case "$(echo "${1#*=}" | tr '[:upper:]' '[:lower:]')" in
          true | yes | 1)
            MULTI=true
            CUDA_VERSION="MULTI"
            if [ "$CUDA_FLAG_SET" == true ]; then
                echo "Ignoring --cuda flag when multi is set to true. Creating definition file for a CUDA Multi distribution."
            fi
            ;;
          false | no | 0)
            MULTI=false
            ;;
          *)
            echo "Invalid value for --multi: ${1#*=}. Allowed values are true/false, yes/no, 1/0." >&2
            exit 1
            ;;
        esac
        ;;
      *)
        echo "Unrecognized flag: $1" >&2
        exit 1
        ;;
    esac
    shift
  done
fi

if [ "$MULTI" == false ]; then
  validate_cuda_version "$NVHPC_VERSION" "$CUDA_VERSION"
fi


output_file="$script_dir/ecwam_nvhpc-${NVHPC_VERSION}_cuda-${CUDA_VERSION}.def"

# Set the template and output files based on the MULTI flag
if [ "$MULTI" == true ]; then
    template_file="$script_dir/singularity_ecwam_multi.def"
    # Replace the template variables with the specified values
    sed -e "s|REPO_DIR|$REPONAME|g" -e "s|NVHPC_VERSION|$NVHPC_VERSION|g" -e "s|CUDA_VERSION|$CUDA_VERSION|g" "$template_file" > "$output_file"
    # Inform the user about the just created definition file
    echo "Singularity definition file for a develop CUDA Multi distribution created at $output_file"
    echo "Provided compiler is nvhpc version $NVHPC_VERSION"
else
    template_file="$script_dir/singularity_ecwam.def"
    # Replace the template variables with the specified values
    sed -e "s|REPO_DIR|$REPONAME|g" -e "s|NVHPC_VERSION|$NVHPC_VERSION|g" -e "s|CUDA_VERSION|$CUDA_VERSION|g" "$template_file" > "$output_file"
    # Inform the user about the just created definition file
    echo "Singularity definition file for a runtime CUDA distribution created at $output_file"
    echo "Provided compiler is nvhpc version $NVHPC_VERSION, cuda version is $CUDA_VERSION"
fi


